/*
 * server.c SPAMNET Network Tester Server
 *
 * Copyright 2003 MontaVista Software
 *
 * Authors: Steven Dake (sdake@mvista.com) initial code
 * 		John Mehaffey (mehaf@mvista.com) added getopt
 * 		Corey Minyard (cminyard@mvista.com) fixes to handle signals during I/O
 *
 * This program is licensed under the MontaVista Software, Inc. License
 * Agreement ("License Agreement"), and is for the purposes of the License
 * Agreement a MontaVista Licensed Deployment Program.  The License requires
 * that you have a valid Product Subscription and Software License Agreement
 * with MontaVista Software, Inc., or are a Named Contact with active access to
 * the MontaVista Zone, or have a Software License Agreement with MontaVista
 * Software, Inc.  This program comes with no warranties other than those
 * provided for in the Product Subscription agreement.  The License Agreement
 * grants you the right to install, modify, and use the program.  You may
 * distribute the object code and scripts for this program, but you have no
 * right to distribute source code for this program.
 */
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/sockios.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <arpa/inet.h>
#include "dataset.h"

#define _GNU_SOURCE
#include <getopt.h>

#define SERVER_BACKLOG 2
#define SERVER_PORT 2000

/*
 * Globals
 */
static int server_port = SERVER_PORT;
static int backlog = SERVER_BACKLOG;
static int client_count = 0;
static int server_count = 0;
static struct sockaddr_in server_addr;
static char *server_ipaddr=0;

char buffer[16384];

unsigned long long total_messages = 0;
unsigned long long total_compared = 0;
unsigned long long total_megs = 0;
unsigned long long one_mb = 1024*1024;

struct connection {
	int active;
};

int connection_entries = 0;
struct connection *connections = 0;

int server_fd;

struct message {
	int dataset;
	int start;
	int size;
};

/*
 * Implementation
 */
static inline void grow_connections_table (int fd)
{
	if (fd + 1 > connection_entries) {
		connections = realloc (connections, (fd + 1) * sizeof (struct connection));
		if (connections == 0) {
			printf ("out of memory, woops\n");
			exit (1);
		}
		memset (&connections[connection_entries], 0,
		(fd - connection_entries + 1) * sizeof (struct connection));
		connection_entries = fd + 1;
	}
}

static int recv_retry (int s, void *msg, size_t len, int flags)
{
	char   *cmsg = (char *)msg;
	int result;
	size_t total = 0;

retry:
	result = recv(s, cmsg, len, flags);
	/*
	 * Handle signals by retrying sends interrupted by signal
	 */
	if (result == -1) {
		if (errno == EINTR) {
			goto retry;
		} else {
			return (-1);
		}
	}

	total += result;

	if ((flags & MSG_WAITALL) && result < len) {
		len -= result;
		cmsg += result;
		goto retry;
	}

	return total;
}

static int send_retry (int s, const void *msg, size_t len, int flags)
{
	int result;
	int total;
	char *cmsg = (char *)msg;
	int retries = 0;

retry:
	if (retries == 5) {
		return (-1);
	}
	result = send (s, cmsg, len, flags);
	if (result == -1) {
		if (errno == EINTR) {
			goto retry;
		} else
	   if (errno == EAGAIN)	{
			retries += 1;
			goto retry;
		}
	}

	total += result;

	if (result < len) {
		len -= result;
		cmsg += result;
		goto retry;
	}
	return total;
}

static inline int process_receive_message (int fd)
{
	int result;
	struct message message;

	result = recv (fd, &message, sizeof (struct message), MSG_WAITALL | MSG_NOSIGNAL);
	if (result != sizeof (struct message)) {
		return (-1);
	}
	
	message.dataset = ntohl (message.dataset);
	message.start = ntohl (message.start);
	message.size = ntohl (message.size);

	result = recv_retry (fd, buffer, message.size, MSG_WAITALL | MSG_NOSIGNAL);
	if (result != message.size) {
		return (-1);
	}
	if (memcmp (buffer, &datasets[message.dataset].dataset[message.start], message.size) != 0) {
		printf ("Erroneous data compare at message %lld\n", total_messages);
		exit (1);
	}

	/*
	 * Bookeeping
	 */
	total_compared += message.size;
	total_messages += 1;
	if ((total_compared / one_mb) > total_megs) {
		total_megs = total_compared / one_mb;
		printf ("Processed %lld messages, %lld MB\n", total_messages, total_megs);
	}

	return (result);
}

static int client (char **argv)
{
	int client_fd;
	struct message message;
	int result;
	int dataset;
	int size;
	int start;
	char process_name[80];

	for (;;) {
		/*
		 * this is a client process, run the client
		 */
		sprintf (process_name, "SPAMNET: Client connecting to %s\n", server_ipaddr);
		strcpy(argv[0], process_name);
	
		/*
		 * Try to connect
		 */
		do {
			printf ("trying to connect\n");
			client_fd = socket (AF_INET, SOCK_STREAM, 0);
			result = connect (client_fd, (struct sockaddr *)&server_addr, sizeof (struct sockaddr_in));
			if (result == -1) {
				sleep (1);
			}
		} while (result == -1);

		sprintf (process_name, "SPAMNET: Client connected to %s\n", server_ipaddr);
		strcpy(argv[0], process_name);

		for (;;) {
			dataset = random() % DATASETS;
			start = random() % datasets[dataset].size;
			size = random() % (datasets[dataset].size - start);
			message.dataset = htonl (dataset);
			message.start = htonl (start);
			message.size = htonl (size);
			result = send_retry (client_fd, &message, sizeof (struct message), MSG_NOSIGNAL);
			if (result == -1) {
				perror ("send");
				break;
			}

			result = send_retry (client_fd, &datasets[dataset].dataset[start], size, MSG_NOSIGNAL);
			if (result == -1) {
				perror ("send");
				break;
			}
		}
	}
}

static int server (char **argv)
{
	struct sockaddr_in in_addr;

	fd_set read_fds;
	int max_fd;
	int result;
	char process_name[80];
	int i;
	int addrlen;
	int new_fd;

	sprintf (process_name, "SPAMNET: Server listening on port %d\n", server_port);
	strcpy(argv[0], process_name);

	/*
	 * Start the server loop
	 */
	for (;;) {
		/*
		 * Wait for incoming data
		 */
		FD_ZERO (&read_fds);
		FD_SET (server_fd, &read_fds);

		max_fd = server_fd;
		for (i = 0; i < connection_entries; i++) {
			if (connections[i].active) {
				FD_SET (i, &read_fds);
				if (i > max_fd) {
					max_fd = i;
				}
			}
		}

retry:
		result = select (max_fd + 1, &read_fds, NULL, NULL, NULL);
		if (result == -1) {
			if (errno == EINTR)
				goto retry;
			perror ("select: ");
			exit (1);
		}

		if (FD_ISSET (server_fd, &read_fds)) {
			addrlen = sizeof (struct sockaddr_in);
			new_fd = accept (server_fd, (struct sockaddr *)&in_addr, &addrlen);
			printf ("connection received from spamnet client %s\n", inet_ntoa (in_addr.sin_addr));
			grow_connections_table (new_fd);
			connections[new_fd].active = 1;
		}

		/*
		 * Check all sockets for data to read, if data, process message
		 */
		for (i = 0; i < connection_entries; i++) {
			if (connections[i].active && FD_ISSET (i, &read_fds)) {
				result = process_receive_message (i);
				/*
				 * Couldn't read any data from socket because it closed
				 */
				if (result == -1) {
					printf ("Closing socket after process_receive_message error\n");
					connections[i].active = 0;
					close (i);
				}
			}
		}
	}
}
static void usage(void)
{
		printf("Options:\n");
		printf("    -i, --server-ip=<IP address of server>\n");
		printf("    -p, --port=server_port\n");
		printf("        use port server_port for clients and servers (default 2000)\n");
		printf("    -s, --server\n");
		printf("        start server\n");
		printf("    -c, --client=<number of clients>\n");
		printf("        start number of clients (default 1)\n");
		printf("    -b, --backlog=listen_backlog\n");
		printf("        max length of pending connections queue (default 2)\n");
}

static struct option long_options[] = {
	{ "server-ip", 1, 0, 'i' },
	{ "port", 1, 0, 'p' },
	{ "server", 0, 0, 's' },
	{ "client", 1, 0, 'c' },
	{ "backlog", 1, 0, 'b' },
	{0, 0, 0, 0}
};
static char short_options[]="i:p:sc:b:";
void get_options(int argc, char** argv)
{
	int opt;
	int result;
	int quit = 0;

	if (argc < 2) {
		usage();
		exit(2);
	}

	while (1) {
		int option_index = 0;

		opt = getopt_long (argc, argv, short_options,
				long_options, &option_index);
		if (opt == -1)
			break;

		switch (opt) {
		case 'i':
			result = inet_pton (AF_INET, optarg, &server_addr.sin_addr);
			if (result == 0) {
				printf ("Invalid IP address %s\n", optarg);
				quit = 1;
			}
			server_addr.sin_family = AF_INET;
			server_addr.sin_port = htons (server_port);
			server_ipaddr = strdup (optarg);
			break;

		case 'p':
			server_port = atoi(optarg);
			if (server_port < 1 || server_port > 0xFFFF) {
				printf ("Invalid server port %s.  Valid range is <1-65535>.\n", optarg);
				quit = 1;
			}
			break;

		case 's':
			server_count = 1;
			break;

		case 'c':
			client_count = atoi (optarg);
			if (client_count < 0 || client_count > 900) {
				printf ("Invalid client count %s.  Valid range is <0-900>.\n", optarg);
				quit = 1;
			}
			break;

		case 'b':
			backlog = atoi(optarg);
			if (backlog < 1 || backlog > 8) {
				printf ("Invalid backlog %s.  Valid range is <1-8>.\n", optarg);
				quit = 1;
			}
			break;

		case '?':
		default:
			usage();
			exit(2);
		}
	}

	if (server_count == 0 && client_count == 0) {
		printf ("No clients or servers specified to be started, exiting.\n");
		quit = 1;
	}
	if (server_ipaddr == 0 && client_count) {
		printf ("If clients are specified, a server ip address must be specified\n");
		quit = 1;
	}
	if (quit == 1) {
			printf ("Fix errors and try again.\n");
		exit (EINVAL);
	}
}

int main (int argc, char **argv, char **envp)
{
	int process;
	int pid;
	struct sockaddr_in in_addr;
	int result;

	printf ("MontaVista Software SPAMNET Network Test Application\n");
	printf ("Copyright (C) 2003 MontaVista Software, Inc.\n");
	get_options (argc, argv);

	printf ("Starting %d servers.\n", server_count);
	printf ("Starting %d clients connecting to address %s.\n", client_count, server_ipaddr);
	if (server_count) {
			/*
			 * Create socket for clients, bind to port, listen for connection
			 */
			server_fd = socket (AF_INET, SOCK_STREAM, 0);
			if (server_fd == -1) {
				perror ("Cannot create client connection socket.");
				exit (1);
			}
			in_addr.sin_family = AF_INET;
			in_addr.sin_addr.s_addr = INADDR_ANY;
			in_addr.sin_port = htons (server_port);
			result = bind (server_fd, (struct sockaddr *)&in_addr, sizeof (struct sockaddr_in));
			if (result) {
				printf ("Could not bind: %s.\n", strerror (errno));
				exit (1);
			}
			listen (server_fd, backlog);
	}

	for (process = 0; process < client_count; process++) {
		pid = fork();
		if (pid == -1) {
			exit (1);
		} else 
		if (pid == 0) {
			/*
			 * Keep fork()ing if this is the master process
			 */
		} else {
			client (argv);
			exit (0);
		}
	}

	/*
	 * Start server, or wait for ctrl-c if no server started
	 */
	if (server_count) {
		printf ("Starting server.\n");
		server (argv);
	} else {
		for (;;) {
			sleep (1);
		}
	}
	return (0);
}
